// SPDX-License-Identifier: GPL-2.0-or-later

/*
 * Copyright (C) 2020 by Nuvoton Technology Corporation
 * Mulin Chao <mlchao@nuvoton.com>
 * Wealian Liao <WHLIAO@nuvoton.com>
 *
 * Copyright (C) 2023 Benjamin VERNOUX 31 Aug 2023 (based on npcx_flash.c)
 */
#include <stdint.h>
#include <string.h>
#include "cx32l003_flash.h"

/*----------------------------------------------------------------------------
 *                             CX32L003 Driver
 *----------------------------------------------------------------------------*/
#ifdef __cplusplus
  #define __I volatile /*!< Defines 'read only' permissions */
#else
  #define __I volatile const /*!< Defines 'read only' permissions */
#endif
#define __IO volatile /*!< Defines 'read / write' permissions */

/**
  * @brief System RCC (RCC)
  */
typedef struct { /*!< (@ 0x40020000) RCC Structure */
__IO uint32_t  HCLKDIV; /*!< (@ 0x00000000) AHB CLK Prescale */
__IO uint32_t  PCLKDIV; /*!< (@ 0x00000004) APB CLK Prescale */
__IO uint32_t  HCLKEN; /*!< (@ 0x00000008) AHB Peripheral Model Clk Enable */
__IO uint32_t  PCLKEN; /*!< (@ 0x0000000C) APB Peripheral Model CLK Enable */
__IO uint32_t  MCOCR; /*!< (@ 0x00000010) Clock Output Control Register */
__I  uint32_t  RESERVED;
__IO uint32_t  RSTCR; /*!< (@ 0x00000018) System Reset Control */
__IO uint32_t  RSTSR; /*!< (@ 0x0000001C) Reset Status */
__IO uint32_t  SYSCLKCR; /*!< (@ 0x00000020) CLK Setting */
__IO uint32_t  SYSCLKSEL; /*!< (@ 0x00000024) System Clock Select */
__IO uint32_t  HIRCCR; /*!< (@ 0x00000028) HIRC Control */
__IO uint32_t  HXTCR; /*!< (@ 0x0000002C) HXT Control */
__IO uint32_t  LIRCCR; /*!< (@ 0x00000030) LIRC Control */
__IO uint32_t  LXTCR; /*!< (@ 0x00000034) LXT Control */
__IO uint32_t  IRQLATENCY; /*!< (@ 0x00000038) M0 IRQ Delay */
__IO uint32_t  STICKCR; /*!< (@ 0x0000003C) SysTick Timer Circle Adjust */
__IO uint32_t  SWDIOCR; /*!< (@ 0x00000040) Endpoint Function Select */
__IO uint32_t  PERIRST; /*!< (@ 0x00000044) Peripheral Model Control */
__IO uint32_t  RTCRST; /*!< (@ 0x00000048) RTC Control */
__I  uint32_t  RESERVED1[5];
__IO uint32_t  UNLOCK; /*!< (@ 0x00000060) Register Protect */
} rcc_typedef; /*!< Size = 100 (0x64) */
#define RCC_BASE 0x40020000UL
#define RCC ((rcc_typedef *) RCC_BASE)

/**
* @brief Independent Watch Dog (IWDG)
*/
typedef struct { /*!< (@ 0x40002400) IWDG Structure */
__IO uint32_t  CMDCR; /*!< (@ 0x00000000) IWDG Control Command Register */
__IO uint32_t  CFGR;  /*!< (@ 0x00000004) IWDG Config Register */
__IO uint32_t  RLOAD; /*!< (@ 0x00000008) Count ReLoad Register */
__IO uint32_t  CNTVAL;/*!< (@ 0x0000000C) Count Value */
__IO uint32_t  SR;    /*!< (@ 0x00000010) IWDG Interrupt Status Register */
__IO uint32_t  INTCLR;/*!< (@ 0x00000014) IWDG Interrupt Status Register */
__IO uint32_t  UNLOCK;/*!< (@ 0x00000018) IWDG register Writer Protect */
} iwdg_typedef; /*!< Size = 28 (0x1c) */
#define IWDG_BASE 0x40002400UL
#define IWDG ((iwdg_typedef *) IWDG_BASE)

/**
  * @brief Flash Controller (FLASH)
  */
typedef struct { /*!< (@ 0x40020400) FLASH Structure */
__IO uint32_t  CR; /*!< (@ 0x00000000) Control Register */
__IO uint32_t  IFR; /*!< (@ 0x00000004) Interrupt flag Register */
__IO uint32_t  ICLR; /*!< (@ 0x00000008) Interrupt Flag Clear Register */
__IO uint32_t  BYPASS; /*!< (@ 0x0000000C) 0X5A5A-0XA5A5 sequence Register */
__IO uint32_t  SLOCK0; /*!< (@ 0x00000010) Sector Write Protect Register0 */
__IO uint32_t  SLOCK1; /*!< (@ 0x00000014) Sector Write Protect Register1 */
__IO uint32_t  ISPCON; /*!< (@ 0x00000018) Flash ISP Control register */
} flash_typedef; /*!< Size = 28 (0x1c) */
#define FLASH_BASE 0x40020400UL
#define FLASH ((flash_typedef *) FLASH_BASE)

void cx32l003_flash_erase_sector(uint32_t page_addr)
{
	uint32_t val;
	/* If IWDG is running, reload it to max value about 27s */
	val = IWDG->CFGR;
	if ((val & 4) == 4) {
		IWDG->UNLOCK = 0x55aa6699;
		IWDG->RLOAD = 0xfffff;
		IWDG->UNLOCK = 0x55aa6698;
	}

	if (page_addr < 0x8000) {
		FLASH->BYPASS = 0x5a5a;
		FLASH->BYPASS = 0xa5a5;
		val = FLASH->SLOCK0;
		FLASH->SLOCK0 = val | 1 << (page_addr >> 10 & 0xff);
	} else if (page_addr < 0x10000) {
		FLASH->BYPASS = 0x5a5a;
		FLASH->BYPASS = 0xa5a5;
		val = FLASH->SLOCK1;
		FLASH->SLOCK1 = val | 1 << ((page_addr - 0x8000) >> 10 & 0xff);
		FLASH->BYPASS = 0x5a5a;
		FLASH->BYPASS = 0;
	}
	FLASH->BYPASS = 0x5a5a;
	FLASH->BYPASS = 0xa5a5;
	FLASH->CR = 2;
	/* Erase flash sector */
	*(uint32_t *)page_addr = 0;

	/* Wait for end of erase "idle state" */
	do {
		val = FLASH->CR;
	} while ((val & 4) == 4);

	FLASH->BYPASS = 0x5a5a;
	FLASH->BYPASS = 0xa5a5;
	val = FLASH->CR;
	FLASH->CR = val & 0xffffff00;
	if (page_addr < 0x8000) {
		FLASH->BYPASS = 0x5a5a;
		FLASH->BYPASS = 0xa5a5;
		val = FLASH->SLOCK0;
		FLASH->SLOCK0 = val & ~(1 << (page_addr >> 10 & 0xff));
	} else if (page_addr < 0x10000) {
		FLASH->BYPASS = 0x5a5a;
		FLASH->BYPASS = 0xa5a5;
		val = FLASH->SLOCK1;
		FLASH->SLOCK1 = val & ~(1 << ((page_addr - 0x8000) >> 10 & 0xff));
		FLASH->BYPASS = 0x5a5a;
		FLASH->BYPASS = 0;
	}
}

void cx32l003_flash_program(uint32_t *flash_addr_dst, int size, uint32_t *ram_data_src)
{
	uint32_t val;

	/* If IWDG is running, reload it to max value about 27s */
	val = IWDG->CFGR;
	if ((val & 4) == 4) {
		IWDG->UNLOCK = 0x55aa6699;
		IWDG->RLOAD = 0xfffff;
		IWDG->UNLOCK = 0x55aa6698;
	}
	for (; size != 0; size = size - 4) {
		if (flash_addr_dst < (uint32_t *)0x8000) {
			FLASH->BYPASS = 0x5a5a;
			FLASH->BYPASS = 0xa5a5;
			val = FLASH->SLOCK0;
			FLASH->SLOCK0 = val | 1 << ((uint32_t)flash_addr_dst >> 10 & 0xff);
		} else if (flash_addr_dst < (uint32_t *)0x10000) {
			FLASH->BYPASS = 0x5a5a;
			FLASH->BYPASS = 0xa5a5;
			val = FLASH->SLOCK1;
			FLASH->SLOCK1 = val | 1 << ((uint32_t)(flash_addr_dst - 0x2000) >> 10 & 0xff);
			FLASH->BYPASS = 0x5a5a;
			FLASH->BYPASS = 0;
		}
		FLASH->BYPASS = 0x5a5a;
		FLASH->BYPASS = 0xa5a5;
		val = FLASH->CR;
		FLASH->CR = val & 0xffffff00;
		FLASH->BYPASS = 0x5a5a;
		FLASH->BYPASS = 0xa5a5;
		val = FLASH->CR;
		FLASH->CR = val | 1;

		/* Program flash */
		*flash_addr_dst = *ram_data_src;

		/* Wait end of program "idle state" */
		do {
			val = FLASH->CR;
		} while ((val & 4) == 4);

		if (flash_addr_dst < (uint32_t *)0x8000) {
			FLASH->BYPASS = 0x5a5a;
			FLASH->BYPASS = 0xa5a5;
			val = FLASH->SLOCK0;
			FLASH->SLOCK0 = val & ~(1 << ((uint32_t)flash_addr_dst >> 10 & 0xff));
		} else if (flash_addr_dst < (uint32_t *)0x10000) {
			FLASH->BYPASS = 0x5a5a;
			FLASH->BYPASS = 0xa5a5;
			val = FLASH->SLOCK1;
			FLASH->SLOCK1 = val & ~(1 << ((uint32_t)(flash_addr_dst - 0x2000) >> 10 & 0xff));
			FLASH->BYPASS = 0x5a5a;
			FLASH->BYPASS = 0;
		}
		flash_addr_dst = flash_addr_dst + 1;
		ram_data_src = ram_data_src + 1;
	}
}

/*----------------------------------------------------------------------------
 *                             flash loader function
 *----------------------------------------------------------------------------*/
uint32_t flashloader_init(struct cx32l003_flash_params *params)
{
	/* Initialize params buffers */
	memset(params, 0, sizeof(struct cx32l003_flash_params));

	return CX32L003_FLASH_STATUS_OK;
}

/*----------------------------------------------------------------------------
 *                                      Functions
 *----------------------------------------------------------------------------*/
/* flashloader parameter structure */
__attribute__ ((section(".buffers.g_cfg")))
volatile struct cx32l003_flash_params g_cfg;
/* data buffer */
__attribute__ ((section(".buffers.g_buf")))
uint8_t g_buf[CX32L003_FLASH_LOADER_BUFFER_SIZE];

static uint32_t val;
static int status;
static uint16_t address;
static uint32_t size;

void init(void)
{
	// RCC Unlock is done in openocd cx32l003.cfg
	/* Flash controller block clock enable */
	val = RCC->HCLKEN;
	RCC->HCLKEN = val | 0x100;

	/* If IWDG is running, reload it to max value about 27s */
	val = IWDG->CFGR;
	if ((val & 4) == 4) {
		IWDG->UNLOCK = 0x55aa6699;
		IWDG->RLOAD = 0xfffff;
		IWDG->UNLOCK = 0x55aa6698;
	}

	/* Wait Flash "idle state" */
	do {
		val = FLASH->CR;
	} while ((val & 4) == 4);
}

int main(void)
{
	init();

	/* set buffer */
	flashloader_init((struct cx32l003_flash_params *)&g_cfg);

	while (1) {
		/* wait command*/
		while (g_cfg.sync == CX32L003_FLASH_LOADER_WAIT) {
			/* If IWDG is running, reload it to max value about 27s */
			val = IWDG->CFGR;
			if ((val & 4) == 4) {
				IWDG->UNLOCK = 0x55aa6699;
				IWDG->RLOAD = 0xfffff;
				IWDG->UNLOCK = 0x55aa6698;
			}
		}

		/* command handler */
		switch (g_cfg.cmd) {
			case CX32L003_FLASH_CMD_GET_FLASH_ID:
				/* Hard coded identifier 0x003203F8 for CX32L003F8 */
				g_buf[0] = 0xF8; /* LSB */
				g_buf[1] = 0x03;
				g_buf[2] = 0x32;
				g_buf[3] = 0x00; /* MSB */
				status = CX32L003_FLASH_STATUS_OK;
				break;
			case CX32L003_FLASH_CMD_ERASE_SECTORS:
				/* Alignment has been checked in upper layer */
				address = g_cfg.addr;
				size = g_cfg.len;
				for (; size > 0; size -= CX32L003_FLASH_ERASE_SIZE,
					address += CX32L003_FLASH_ERASE_SIZE) {
					cx32l003_flash_erase_sector(address);
				}
				status = CX32L003_FLASH_STATUS_OK;
				break;
			case CX32L003_FLASH_CMD_ERASE_ALL:
				address = 0x0;
				size = 64 * 1024; /* 64KB */
				for (; size > 0; size -= CX32L003_FLASH_ERASE_SIZE,
					address += CX32L003_FLASH_ERASE_SIZE) {
					cx32l003_flash_erase_sector(address);
				}
				status = CX32L003_FLASH_STATUS_OK;
				break;
			case CX32L003_FLASH_CMD_PROGRAM:
				cx32l003_flash_program((uint32_t *)g_cfg.addr, g_cfg.len, (uint32_t *)g_buf);
				status = CX32L003_FLASH_STATUS_OK;
				break;
			default:
				status = CX32L003_FLASH_STATUS_FAILED_UNKNOWN_COMMAND;
				break;
		}

		/* clear & set result for next command */
		if (status != CX32L003_FLASH_STATUS_OK) {
			g_cfg.sync = status;
			// Fatal error infinite loop
			while (1)
				;
		} else {
			g_cfg.sync = CX32L003_FLASH_LOADER_WAIT;
		}
	}

	return 0;
}
