// SPDX-License-Identifier: GPL-2.0-or-later

/*
 * Copyright (C) 2020 by Nuvoton Technology Corporation
 * Mulin Chao <mlchao@nuvoton.com>
 * Wealian Liao <WHLIAO@nuvoton.com>
 *
 * Copyright (C) 2023 Benjamin VERNOUX 02 Sep 2023 (based on npcx.c)
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "imp.h"
#include <helper/binarybuffer.h>
#include <helper/time_support.h>
#include <target/armv7m.h>
#include "../../../contrib/loaders/flash/cx32l003/cx32l003_flash.h"

/* CX32L003 flash loader */
static const uint8_t cx32l003_algo[] = {
#include "../../../contrib/loaders/flash/cx32l003/cx32l003_algo.inc"
};

#define CX32L003_FLASH_TIMEOUT_MS 10000 // Timeout 10s max
#define CX32L003_FLASH_BASE_ADDR 0x00000000
#define CX32L003_FLASH_SECTOR_LENGTH (CX32L003_FLASH_ERASE_SIZE)

/* flash list */
enum cx32l003_flash_device_index {
	CX32L003_FLASH_64KB = 0,
	CX32L003_FLASH_UNKNOWN,
};

struct cx32l003_flash_bank {
	const char *family_name;
	uint32_t sector_length;
	bool probed;
	enum cx32l003_flash_device_index flash;
	struct working_area *working_area;
	struct armv7m_algorithm armv7m_info;
	const uint8_t *algo_code;
	uint32_t algo_size;
	uint32_t algo_working_size;
	uint32_t buffer_addr;
	uint32_t params_addr;
};

struct cx32l003_flash_info {
	char *name;
	uint32_t id;
	uint32_t size;
};

static const struct cx32l003_flash_info flash_info[] = {
	[CX32L003_FLASH_64KB] = {
		.name = "64KB Flash",
		.id = 0x003203F8,
		.size = 64 * 1024,
	},
	[CX32L003_FLASH_UNKNOWN] = {
		.name = "Unknown Flash",
		.size = 0xFFFFFFFF,
	},
};

static int cx32l003_init(struct flash_bank *bank)
{
	struct target *target = bank->target;
	struct cx32l003_flash_bank *cx32l003_bank = bank->driver_priv;

	/* Check for working area to use for flash helper algorithm */
	target_free_working_area(target, cx32l003_bank->working_area);
	cx32l003_bank->working_area = NULL;

	int retval = target_alloc_working_area(target, cx32l003_bank->algo_working_size,
				&cx32l003_bank->working_area);
	if (retval != ERROR_OK)
		return retval;

	/* Confirm the defined working address is the area we need to use */
	if (cx32l003_bank->working_area->address != CX32L003_FLASH_LOADER_WORKING_ADDR) {
		LOG_ERROR("%s: Invalid working address", cx32l003_bank->family_name);
		LOG_INFO("Hint: Use '-work-area-phys 0x%" PRIx32 "' in your target configuration",
			CX32L003_FLASH_LOADER_WORKING_ADDR);
		target_free_working_area(target, cx32l003_bank->working_area);
		cx32l003_bank->working_area = NULL;
		return ERROR_TARGET_RESOURCE_NOT_AVAILABLE;
	}

	/* Write flash helper algorithm into target memory */
	retval = target_write_buffer(target, CX32L003_FLASH_LOADER_PROGRAM_ADDR,
				cx32l003_bank->algo_size, cx32l003_bank->algo_code);
	if (retval != ERROR_OK) {
		LOG_ERROR("%s: Failed to load flash helper algorithm",
			cx32l003_bank->family_name);
		target_free_working_area(target, cx32l003_bank->working_area);
		cx32l003_bank->working_area = NULL;
		return retval;
	}

	/* Initialize the ARMv7 specific info to run the algorithm */
	cx32l003_bank->armv7m_info.common_magic = ARMV7M_COMMON_MAGIC;
	cx32l003_bank->armv7m_info.core_mode = ARM_MODE_THREAD;

	/* Begin executing the flash helper algorithm */
	retval = target_start_algorithm(target, 0, NULL, 0, NULL,
				CX32L003_FLASH_LOADER_PROGRAM_ADDR, 0,
				&cx32l003_bank->armv7m_info);
	if (retval != ERROR_OK) {
		LOG_ERROR("%s: Failed to start flash helper algorithm",
			cx32l003_bank->family_name);
		target_free_working_area(target, cx32l003_bank->working_area);
		cx32l003_bank->working_area = NULL;
		return retval;
	}
	/*
	 * At this point, the algorithm is running on the target and
	 * ready to receive commands and data to flash the target
	 */
	return retval;
}

static int cx32l003_quit(struct flash_bank *bank)
{
	struct target *target = bank->target;
	struct cx32l003_flash_bank *cx32l003_bank = bank->driver_priv;

	/* Regardless of the algo's status, attempt to halt the target */
	(void)target_halt(target);

	/* Now confirm target halted and clean up from flash helper algorithm */
	int retval = target_wait_algorithm(target, 0, NULL, 0, NULL, 0,
					CX32L003_FLASH_TIMEOUT_MS, &cx32l003_bank->armv7m_info);

	target_free_working_area(target, cx32l003_bank->working_area);
	cx32l003_bank->working_area = NULL;

	return retval;
}

static int cx32l003_wait_algo_done(struct flash_bank *bank, uint32_t params_addr)
{
	struct target *target = bank->target;
	struct cx32l003_flash_bank *cx32l003_bank = bank->driver_priv;
	uint32_t status_addr = params_addr + offsetof(struct cx32l003_flash_params, sync);
	uint32_t status;
	int64_t start_ms = timeval_ms();

	do {
		int retval = target_read_u32(target, status_addr, &status);
		if (retval != ERROR_OK)
			return retval;

		keep_alive();

		int64_t elapsed_ms = timeval_ms() - start_ms;
		if (elapsed_ms > CX32L003_FLASH_TIMEOUT_MS)
			break;
	} while (status == CX32L003_FLASH_LOADER_EXECUTE);

	if (status != CX32L003_FLASH_LOADER_WAIT) {
		LOG_ERROR("%s: Flash operation failed, status=0x%" PRIx32,
				cx32l003_bank->family_name,
				status);
		return ERROR_FAIL;
	}

	return ERROR_OK;
}

#if 0
static enum cx32l003_flash_device_index cx32l003_get_flash_id(struct flash_bank *bank, uint32_t *flash_id)
{
	struct target *target = bank->target;
	struct cx32l003_flash_bank *cx32l003_bank = bank->driver_priv;
	struct cx32l003_flash_params algo_params;

	if (target->state != TARGET_HALTED) {
		LOG_ERROR("Target not halted");
		return ERROR_TARGET_NOT_HALTED;
	}

	int retval = cx32l003_init(bank);
	if (retval != ERROR_OK)
		return retval;

	/* Set up algorithm parameters for get flash ID command */
	target_buffer_set_u32(target, (uint8_t *)&algo_params.cmd, CX32L003_FLASH_CMD_GET_FLASH_ID);
	target_buffer_set_u32(target, (uint8_t *)&algo_params.sync, CX32L003_FLASH_LOADER_WAIT);

	/* Issue flash helper algorithm parameters for get flash ID */
	retval = target_write_buffer(target, cx32l003_bank->params_addr,
				sizeof(algo_params), (uint8_t *)&algo_params);
	if (retval != ERROR_OK) {
		(void)cx32l003_quit(bank);
		return retval;
	}

	target_buffer_set_u32(target, (uint8_t *)&algo_params.sync, CX32L003_FLASH_LOADER_EXECUTE);
	retval = target_write_buffer(target, cx32l003_bank->params_addr,
				sizeof(algo_params), (uint8_t *)&algo_params);

	/* If no error, wait for finishing */
	if (retval == ERROR_OK) {
		retval = cx32l003_wait_algo_done(bank, cx32l003_bank->params_addr);
		if (retval == ERROR_OK)
			target_read_u32(target, CX32L003_FLASH_LOADER_BUFFER_ADDR, flash_id);
	}

	/* Regardless of errors, try to close down algo */
	(void)cx32l003_quit(bank);

	return retval;
}

static int cx32l003_get_flash(uint32_t flash_id)
{
	for (uint32_t i = 0; i < ARRAY_SIZE(flash_info) - 1; i++) {
		if (flash_info[i].id == flash_id)
			return i;
	}
	return CX32L003_FLASH_UNKNOWN;
}
#endif

static int cx32l003_probe(struct flash_bank *bank)
{
#if 0
	struct cx32l003_flash_bank *cx32l003_bank = bank->driver_priv;
	uint32_t sector_length = CX32L003_FLASH_ERASE_SIZE;
	uint32_t flash_id;

	/* Set up appropriate flash helper algorithm */
	cx32l003_bank->algo_code = cx32l003_algo;
	cx32l003_bank->algo_size = sizeof(cx32l003_algo);
	cx32l003_bank->algo_working_size = CX32L003_FLASH_LOADER_PARAMS_SIZE +
					CX32L003_FLASH_LOADER_BUFFER_SIZE +
					CX32L003_FLASH_LOADER_PROGRAM_SIZE;
	cx32l003_bank->buffer_addr = CX32L003_FLASH_LOADER_BUFFER_ADDR;
	cx32l003_bank->params_addr = CX32L003_FLASH_LOADER_PARAMS_ADDR;

	int retval = cx32l003_get_flash_id(bank, &flash_id);
	if (retval != ERROR_OK)
		return retval;

	cx32l003_bank->flash = cx32l003_get_flash(flash_id);
	if (cx32l003_bank->flash == CX32L003_FLASH_UNKNOWN) {
		LOG_ERROR("cx32l003_bank->flash CX32L003_FLASH_UNKNOWN id=0x%08X bank id=%d", flash_id, cx32l003_bank->flash);
		return ERROR_FAIL;
	}

	unsigned int num_sectors = flash_info[cx32l003_bank->flash].size / sector_length;
	LOG_INFO("cx32l003_get_flash_id size=%d bytes, sector_length=%d bytes num_sectors=%d",
			flash_info[cx32l003_bank->flash].size, sector_length, num_sectors);

	bank->sectors = calloc(num_sectors, sizeof(struct flash_sector));
	if (!bank->sectors) {
		LOG_ERROR("Out of memory");
		return ERROR_FAIL;
	}

	bank->base = CX32L003_FLASH_BASE_ADDR;
	bank->num_sectors = num_sectors;
	bank->size = num_sectors * sector_length;
	bank->write_start_alignment = 4;
	bank->write_end_alignment = 4;
	cx32l003_bank->sector_length = sector_length;

	for (unsigned int i = 0; i < num_sectors; i++) {
		bank->sectors[i].offset = i * sector_length;
		bank->sectors[i].size = sector_length;
		bank->sectors[i].is_erased = -1;
		bank->sectors[i].is_protected = 0;
	}

	/* We've successfully determined the stats on the flash bank */
	cx32l003_bank->probed = true;

	/* If we fall through to here, then all went well */
	return ERROR_OK;
#else
	return ERROR_OK;
#endif
}

static int cx32l003_auto_probe(struct flash_bank *bank)
{
#if 0
	struct cx32l003_flash_bank *cx32l003_bank = bank->driver_priv;
	int retval = ERROR_OK;

	if (!cx32l003_bank->probed)
		retval = cx32l003_probe(bank);

	return retval;
#else
	return ERROR_OK;
#endif
}

FLASH_BANK_COMMAND_HANDLER(cx32l003_flash_bank_command)
{
	struct cx32l003_flash_bank *cx32l003_bank;

	if (CMD_ARGC < 6)
		return ERROR_COMMAND_SYNTAX_ERROR;

	cx32l003_bank = calloc(1, sizeof(struct cx32l003_flash_bank));
	if (!cx32l003_bank) {
		LOG_ERROR("Out of memory");
		return ERROR_FAIL;
	}

	/* Initialize private flash information */
	cx32l003_bank->family_name = "cx32l003";
	cx32l003_bank->sector_length = CX32L003_FLASH_ERASE_SIZE;

	/* Set up appropriate flash helper algorithm */
	cx32l003_bank->algo_code = cx32l003_algo;
	cx32l003_bank->algo_size = sizeof(cx32l003_algo);
	cx32l003_bank->algo_working_size = CX32L003_FLASH_LOADER_PARAMS_SIZE +
					CX32L003_FLASH_LOADER_BUFFER_SIZE +
					CX32L003_FLASH_LOADER_PROGRAM_SIZE;
	cx32l003_bank->buffer_addr = CX32L003_FLASH_LOADER_BUFFER_ADDR;
	cx32l003_bank->params_addr = CX32L003_FLASH_LOADER_PARAMS_ADDR;

	/* Finish initialization of bank */
	bank->driver_priv = cx32l003_bank;
	bank->next = NULL;

	/* bank init hard coded */
	/* CX32L003F8 64KB Flash (128 sectors * 512 bytes) */
	#define CX32L003_FLASH_PAGE_SIZE	512
	#define CX32L003_FLASH_PAGES		128
	free(bank->sectors);
	bank->write_start_alignment = 4;
	bank->write_end_alignment = 4;
	bank->size = CX32L003_FLASH_PAGE_SIZE * CX32L003_FLASH_PAGES;

	bank->num_sectors = CX32L003_FLASH_PAGES;
	bank->sectors = alloc_block_array(0, CX32L003_FLASH_PAGE_SIZE, CX32L003_FLASH_PAGES);
	if (!bank->sectors)
		return ERROR_FAIL;

	for (unsigned int i = 0; i < bank->num_sectors; i++)
		bank->sectors[i].is_protected = 0;


	return ERROR_OK;
}

static int cx32l003_chip_erase(struct flash_bank *bank)
{
	struct target *target = bank->target;
	struct cx32l003_flash_bank *cx32l003_bank = bank->driver_priv;
	struct cx32l003_flash_params algo_params;

	if (target->state != TARGET_HALTED) {
		LOG_ERROR("Target not halted");
		return ERROR_TARGET_NOT_HALTED;
	}

	/* Make sure we've probed the flash to get the device and size */
	int retval = cx32l003_auto_probe(bank);
	if (retval != ERROR_OK)
		return retval;

	retval = cx32l003_init(bank);
	if (retval != ERROR_OK)
		return retval;

	/* Set up algorithm parameters for chip erase command */
	target_buffer_set_u32(target, (uint8_t *)&algo_params.cmd, CX32L003_FLASH_CMD_ERASE_ALL);
	target_buffer_set_u32(target, (uint8_t *)&algo_params.sync, CX32L003_FLASH_LOADER_WAIT);

	/* Set algorithm parameters */
	retval = target_write_buffer(target, cx32l003_bank->params_addr,
				sizeof(algo_params), (uint8_t *)&algo_params);
	if (retval != ERROR_OK) {
		(void)cx32l003_quit(bank);
		return retval;
	}

	/* Issue flash helper algorithm parameters for chip erase */
	target_buffer_set_u32(target, (uint8_t *)&algo_params.sync, CX32L003_FLASH_LOADER_EXECUTE);
	retval = target_write_buffer(target, cx32l003_bank->params_addr,
				sizeof(algo_params), (uint8_t *)&algo_params);

	/* If no error, wait for chip erase finish */
	if (retval == ERROR_OK)
		retval = cx32l003_wait_algo_done(bank, cx32l003_bank->params_addr);

	/* Regardless of errors, try to close down algo */
	(void)cx32l003_quit(bank);

	return retval;
}

static int cx32l003_erase(struct flash_bank *bank, unsigned int first,
		unsigned int last)
{
	struct target *target = bank->target;
	struct cx32l003_flash_bank *cx32l003_bank = bank->driver_priv;
	struct cx32l003_flash_params algo_params;
	int retval;
	long long start_ms;
	int elapsed_ms;

	if (target->state != TARGET_HALTED) {
		LOG_ERROR("Target not halted");
		return ERROR_TARGET_NOT_HALTED;
	}

	if (first == 0 && (last == (bank->num_sectors - 1))) {
		/* Request chip erase */
		LOG_INFO("cx32l003_mass_erase");
		start_ms = timeval_ms();
		retval = cx32l003_chip_erase(bank);
		elapsed_ms = (int)(timeval_ms() - start_ms);
		LOG_INFO("cx32l003_mass_erase retval=%d elapsed_ms=%d", retval, elapsed_ms);
		if (retval == ERROR_OK)
			LOG_INFO("cx32l003 mass erase complete");
		else
			LOG_INFO("cx32l003 mass erase failed");

		return retval;
	}

	uint32_t address = first * cx32l003_bank->sector_length;
	uint32_t length = (last - first + 1) * cx32l003_bank->sector_length;

	LOG_INFO("cx32l003_erase first=%d last=%d", first, last);
	start_ms = timeval_ms();

	/* Make sure we've probed the flash to get the device and size */
	retval = cx32l003_auto_probe(bank);
	if (retval != ERROR_OK)
		return retval;

	retval = cx32l003_init(bank);
	if (retval != ERROR_OK)
		return retval;

	/* Set up algorithm parameters for erase command */
	target_buffer_set_u32(target, (uint8_t *)&algo_params.addr, address);
	target_buffer_set_u32(target, (uint8_t *)&algo_params.len, length);
	target_buffer_set_u32(target, (uint8_t *)&algo_params.cmd, CX32L003_FLASH_CMD_ERASE_SECTORS);
	target_buffer_set_u32(target, (uint8_t *)&algo_params.sync, CX32L003_FLASH_LOADER_WAIT);

	/* Set algorithm parameters */
	retval = target_write_buffer(target, cx32l003_bank->params_addr,
				sizeof(algo_params), (uint8_t *)&algo_params);
	if (retval != ERROR_OK) {
		(void)cx32l003_quit(bank);
		return retval;
	}

	/* Issue flash helper algorithm parameters for erase */
	target_buffer_set_u32(target, (uint8_t *)&algo_params.sync, CX32L003_FLASH_LOADER_EXECUTE);
	retval = target_write_buffer(target, cx32l003_bank->params_addr,
				sizeof(algo_params), (uint8_t *)&algo_params);

	/* If no error, wait for erase to finish */
	if (retval == ERROR_OK)
		retval = cx32l003_wait_algo_done(bank, cx32l003_bank->params_addr);

	/* Regardless of errors, try to close down algo */
	(void)cx32l003_quit(bank);

	elapsed_ms = (int)(timeval_ms() - start_ms);
	LOG_INFO("cx32l003_erase retval=%d elapsed_ms=%d", retval, elapsed_ms);

	return retval;
}

static int cx32l003_write(struct flash_bank *bank, const uint8_t *buffer,
	uint32_t offset, uint32_t count)
{
	struct target *target = bank->target;
	int retval;
	long long start_ms;
	int elapsed_ms;
	struct cx32l003_flash_bank *cx32l003_bank = bank->driver_priv;
	struct cx32l003_flash_params algo_params;

	if (target->state != TARGET_HALTED) {
		LOG_ERROR("Target not halted");
		return ERROR_TARGET_NOT_HALTED;
	}

	LOG_INFO("cx32l003_write buffer=%p offset=0x%08X count=%d", buffer, offset, count);
	start_ms = timeval_ms();
	
	/* Make sure we've probed the flash to get the device and size */
	retval = cx32l003_auto_probe(bank);
	if (retval != ERROR_OK)
		return retval;

	retval = cx32l003_init(bank);
	if (retval != ERROR_OK)
		return retval;

	/* Initialize algorithm parameters to default values */
	target_buffer_set_u32(target, (uint8_t *)&algo_params.cmd, CX32L003_FLASH_CMD_PROGRAM);

	uint32_t address = offset;

	while (count > 0) {
		uint32_t size = (count > CX32L003_FLASH_LOADER_BUFFER_SIZE) ?
							CX32L003_FLASH_LOADER_BUFFER_SIZE : count;

		/* Put the data into buffer */
		retval = target_write_buffer(target, cx32l003_bank->buffer_addr,
					size, buffer);
		if (retval != ERROR_OK) {
			LOG_ERROR("Unable to write data to target memory");
			break;
		}

		/* Update algo parameters for flash write */
		target_buffer_set_u32(target, (uint8_t *)&algo_params.addr, address);
		target_buffer_set_u32(target, (uint8_t *)&algo_params.len, size);
		target_buffer_set_u32(target, (uint8_t *)&algo_params.sync, CX32L003_FLASH_LOADER_WAIT);

		/* Set algorithm parameters */
		retval = target_write_buffer(target, cx32l003_bank->params_addr,
				sizeof(algo_params), (uint8_t *)&algo_params);
		if (retval != ERROR_OK)
			break;

		/* Issue flash helper algorithm parameters for flash write */
		target_buffer_set_u32(target, (uint8_t *)&algo_params.sync, CX32L003_FLASH_LOADER_EXECUTE);
		retval = target_write_buffer(target, cx32l003_bank->params_addr,
				sizeof(algo_params), (uint8_t *)&algo_params);
		if (retval != ERROR_OK)
			break;

		/* Wait for flash write finish */
		retval = cx32l003_wait_algo_done(bank, cx32l003_bank->params_addr);
		if (retval != ERROR_OK)
			break;

		count -= size;
		buffer += size;
		address += size;
	}

	/* Regardless of errors, try to close down algo */
	(void)cx32l003_quit(bank);

	elapsed_ms = (int)(timeval_ms() - start_ms);
	LOG_INFO("cx32l003_write retval=%d elapsed_ms=%d", retval, elapsed_ms);

	return retval;
}

static int cx32l003_info(struct flash_bank *bank, struct command_invocation *cmd)
{
	struct cx32l003_flash_bank *cx32l003_bank = bank->driver_priv;

	command_print_sameline(cmd, "%s flash: %s\n",
					cx32l003_bank->family_name,
					flash_info[cx32l003_bank->flash].name);

	return ERROR_OK;
}

const struct flash_driver cx32l003_flash = {
	.name = "cx32l003",
	.flash_bank_command = cx32l003_flash_bank_command,
	.erase = cx32l003_erase,
	.write = cx32l003_write,
	.read = default_flash_read,
	.probe = cx32l003_probe,
	.auto_probe = cx32l003_auto_probe,
	.erase_check = default_flash_blank_check,
	.info = cx32l003_info,
	.free_driver_priv = default_flash_free_driver_priv,
};
